// 
//! @file CHyperLink.cpp
//! HyperLink static control. Will open the default browser with the given URL
//! when the user clicks on the link.
//! @copyright 1992 - 2011 Dennis Robinson (http://www.menasoft.com)
//

#include "StdAfx.h"
#include "CHyperLink.h"

#ifdef _MFC_VER
#include "../System/COSProcess.h"
#include "../System/CAppRes.h"
#include "../../GrayCore/COSModule.h"

#include <atlconv.h>    // for Unicode conversion - requires #include <afxdisp.h> // MFC OLE automation classes

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define TOOLTIP_ID 1

//*******************************************
// CHyperLink

CHyperLink::CHyperLink()
	: m_hLinkCursor(WINHANDLE_NULL)				// No cursor as yet
    , m_crLinkColour(RGB(0,0,238))		// Blue
    , m_crVisitedColour(RGB(85,26,139))	// Purple
    , m_crHoverColour(RGB(255,0,0))		// Red
    , m_bOverControl(false)             // Cursor not yet over control
    , m_bVisited(false)					// Hasn't been visited yet.
    , m_nUnderline( UL_Hover)            // Underline the link?
    , m_bAdjustToFit(true)              // Resize the window to fit the text?
    , m_nTimerID(100)
{
}

CHyperLink::~CHyperLink()
{
    m_UnderlineFont.DeleteObject();
}

//*******************************************
// CHyperLink overrides

BOOL CHyperLink::DestroyWindow() 
{
    KillTimer(m_nTimerID);
	return SUPER_t::DestroyWindow();
}

BOOL CHyperLink::PreTranslateMessage(MSG* pMsg) 
{
    m_ToolTip.RelayEvent(pMsg);
    return SUPER_t::PreTranslateMessage(pMsg);
}

void CHyperLink::PreSubclassWindow() 
{
    // We want to get mouse clicks via STN_CLICKED
    DWORD dwStyle = GetStyle();
    ::SetWindowLong(GetSafeHwnd(), GWL_STYLE, dwStyle | SS_NOTIFY);
    
    // Set the URL as the window text
    if (m_strURL.IsEmpty())
	{
        GetWindowText(m_strURL);
	}

    // Check that the window text isn't empty. If it is, set it as the URL.
    CGString strWndText;
    GetWindowText(strWndText);
    if (strWndText.IsEmpty()) 
    {
        ASSERT(!m_strURL.IsEmpty());    // Window and URL both empty. DUH!
        SetWindowText(m_strURL);
    }

	CFont* pFont = GetFont();
	if (pFont==nullptr)
	{
		HFONT hFont = (HFONT)GetStockObject(DEFAULT_GUI_FONT);
		if (hFont == WINHANDLE_NULL)
			hFont = (HFONT) GetStockObject(ANSI_VAR_FONT);
		if (hFont != WINHANDLE_NULL)
			pFont = CFont::FromHandle(hFont);
	}
	ASSERT(pFont->GetSafeHandle()!=WINHANDLE_NULL);

    // Create the underline font
    CWinFontCfg fontcfg;
    pFont->GetLogFont(&fontcfg);
	m_StdFont.CreateFontIndirect(&fontcfg);
    fontcfg.lfUnderline = (BYTE) true;
    m_UnderlineFont.CreateFontIndirect(&fontcfg);

    PositionWindow();        // Adjust size of window to fit URL if necessary
    SetDefaultCursor();      // Try and load up a "hand" cursor
    SetUnderline();

    // Create the tooltip
    CRect rect; 
    GetClientRect(rect);
    m_ToolTip.Create(this);
    m_ToolTip.AddTool(this, m_strURL, rect, TOOLTIP_ID);

    SUPER_t::PreSubclassWindow();
}

BEGIN_MESSAGE_MAP(CHyperLink, CStatic)
    //{{AFX_MSG_MAP(CHyperLink)
    ON_WM_CTLCOLOR_REFLECT()
    ON_WM_SETCURSOR()
    ON_WM_MOUSEMOVE()
	ON_WM_TIMER()
    ON_CONTROL_REFLECT(STN_CLICKED, OnClicked)
	ON_WM_ERASEBKGND()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

//*******************************************
// CHyperLink message handlers

void CHyperLink::OnClicked()
{
    m_bOverControl = false;
    HRESULT hRes = COSProcessX::ShellExec( nullptr, m_strURL, nullptr, SW_SHOW); // from HINSTANCE
    m_bVisited = SUCCEEDED(hRes);
    if (!m_bVisited)
    {
        MessageBeep(MB_ICONEXCLAMATION);     // Unable to follow link
        ReportError(hRes);
    }
    else 
	{
        SetVisited();                        // Repaint to show visited colour
	}
}

HBRUSH CHyperLink::CtlColor(CDC* pDC, UINT nCtlColor) 
{
    ASSERT(nCtlColor == CTLCOLOR_STATIC);

    if (m_bOverControl)
        pDC->SetTextColor(m_crHoverColour);
    else if (m_bVisited)
        pDC->SetTextColor(m_crVisitedColour);
    else
        pDC->SetTextColor(m_crLinkColour);

    // transparent text.
    pDC->SetBkMode(TRANSPARENT);
    return (HBRUSH)GetStockObject(NULL_BRUSH);
}

void CHyperLink::OnMouseMove(UINT nFlags, CPoint point) 
{
    if (!m_bOverControl)        // Cursor has just moved over control
    {
        m_bOverControl = true;

        if (m_nUnderline == UL_Hover)
            SetFont(&m_UnderlineFont);
        Invalidate();

        SetTimer(m_nTimerID, 100, nullptr);
    }
    SUPER_t::OnMouseMove(nFlags, point);
}

void CHyperLink::OnTimer(UINT_PTR nIDEvent) 
{
    CPoint p(GetMessagePos());
    ScreenToClient(&p);

    CRect rect;
    GetClientRect(rect);
    if (!rect.PtInRect(p))
    {
        m_bOverControl = false;
        KillTimer(m_nTimerID);

        if (m_nUnderline != UL_Always)
            SetFont(&m_StdFont);
        rect.bottom+=10;
        InvalidateRect(rect);
    }
    
	SUPER_t::OnTimer(nIDEvent);
}

BOOL CHyperLink::OnSetCursor(CWnd* /*pWnd*/, UINT /*nHitTest*/, UINT /*message*/) 
{
    if (m_hLinkCursor)
    {
        ::SetCursor(m_hLinkCursor);
        return true;
    }
    return false;
}

BOOL CHyperLink::OnEraseBkgnd(CDC* pDC) 
{
    CRect rect;
    GetClientRect(rect);
    pDC->FillSolidRect(rect, ::GetSysColor(COLOR_3DFACE));
    return true;
}

//*******************************************
// CHyperLink operations

void CHyperLink::SetURL(CGString strURL)
{
    m_strURL = strURL;
    if (::IsWindow(GetSafeHwnd())) 
	{
        PositionWindow();
        m_ToolTip.UpdateTipText(strURL, this, TOOLTIP_ID);
    }
}

void CHyperLink::SetColours(COLORREF crLinkColour, COLORREF crVisitedColour,
                            COLORREF crHoverColour /* = -1 */) 
{ 
    m_crLinkColour    = crLinkColour; 
    m_crVisitedColour = crVisitedColour;

	if (crHoverColour == -1)
		m_crHoverColour = ::GetSysColor(COLOR_HIGHLIGHT);
	else
		m_crHoverColour = crHoverColour;

	if ( ::IsWindow( this->m_hWnd ))
        Invalidate(); 
}

void CHyperLink::SetVisited(BOOL bVisited /* = true */) 
{ 
    m_bVisited = bVisited; 

	if ( ::IsWindow( this->m_hWnd ))
        Invalidate(); 
}

void CHyperLink::SetLinkCursor(HCURSOR hCursor)
{ 
    m_hLinkCursor = hCursor;
    if (m_hLinkCursor == WINHANDLE_NULL)
        SetDefaultCursor();
}

void CHyperLink::SetUnderline( UnderLine_TYPE nUnderline /*=UL_Hover*/)
{
    if (m_nUnderline == nUnderline)
        return;

    if (::IsWindow(GetSafeHwnd()))
    {
        if (nUnderline == UL_Always)
            SetFont(&m_UnderlineFont);
        else
            SetFont(&m_StdFont);

        Invalidate(); 
    }

    m_nUnderline = nUnderline;
}

void CHyperLink::SetAutoSize(BOOL bAutoSize /* = true */)
{
    m_bAdjustToFit = bAutoSize;

    if (::IsWindow(GetSafeHwnd()))
        PositionWindow();
}

void CHyperLink::PositionWindow()
{
	// Move and resize the window so that the window is the same size
	// as the hyperlink text. This stops the hyperlink cursor being active
	// when it is not directly over the text. If the text is left justified
	// then the window is merely shrunk, but if it is centred or right
	// justified then the window will have to be moved as well.

    if (!::IsWindow(GetSafeHwnd()) || !m_bAdjustToFit) 
        return;

    // Get the current window position
    CRect WndRect;
	CRect ClientRect;
    GetWindowRect(WndRect);
    GetClientRect(ClientRect);

    ClientToScreen(ClientRect);

    CWnd* pParent = GetParent();
    if (pParent!=nullptr)
    {
        pParent->ScreenToClient(WndRect);
        pParent->ScreenToClient(ClientRect);
    }

    // Get the size of the window text
    CString strWndText;
    GetWindowText(strWndText);

    CDC* pDC = GetDC();
    CFont* pOldFont = pDC->SelectObject(&m_UnderlineFont);
    CSize Extent = pDC->GetTextExtent(strWndText);
    pDC->SelectObject(pOldFont);
    ReleaseDC(pDC);

    // Adjust for window borders
    Extent.cx += WndRect.Width() - ClientRect.Width(); 
    Extent.cy += WndRect.Height() - ClientRect.Height(); 

    // Get the text justification via the window style
    DWORD dwStyle = GetStyle();

    // Recalc the window size and position based on the text justification
    if (dwStyle & SS_CENTERIMAGE)
        WndRect.DeflateRect(0, (WndRect.Height() - Extent.cy)/2);
    else
        WndRect.bottom = WndRect.top + Extent.cy;

    if (dwStyle & SS_CENTER)   
        WndRect.DeflateRect((WndRect.Width() - Extent.cx)/2, 0);
    else if (dwStyle & SS_RIGHT) 
        WndRect.left  = WndRect.right - Extent.cx;
    else // SS_LEFT = 0, so we can't test for it explicitly 
        WndRect.right = WndRect.left + Extent.cx;

    // Move the window
    SetWindowPos( WINHANDLE_NULL, WndRect.left, WndRect.top, WndRect.Width(), WndRect.Height(), SWP_NOZORDER);
}

//*******************************************
// CHyperLink implementation

void CHyperLink::SetDefaultCursor()
{
	// The following appeared in Paul DiLascia's Jan 1998 MSJ articles.
	// It loads a "hand" cursor from the winhlp32.exe module
    if (m_hLinkCursor != WINHANDLE_NULL)                // No cursor handle - load our own
		return;

	// Get the windows directory
    _TCHAR szWndDir[ _MAX_PATH ];
	StrLen_t iLen = ::GetWindowsDirectory( szWndDir, _countof(szWndDir) );
	StrX::CopyLen( szWndDir+iLen, _T(FILESTR_DirSep) _T("winhlp32.exe"), _countof(szWndDir)-iLen );

    // This retrieves cursor #106 from winhlp32.exe, which is a hand pointer

	COSModule Module;
	HRESULT hRes = Module.LoadModule(szWndDir,COSModule::c_LOADFLAG_FIND);
    if ( FAILED(hRes)) 
		return;

	HCURSOR hHandCursor = ::LoadCursor(Module, CAppRes::MakeIntResource(106));
	if (hHandCursor==WINHANDLE_NULL)
		return;

	m_hLinkCursor = CopyCursor(hHandCursor);	// macro ((HCURSOR)CopyIcon((HICON)(pcur)))
}

void CHyperLink::ReportError(HRESULT hRes)
{
	CGString sErr = "Unable to open hyperlink:\n\n";
	sErr += CGString::GetErrorMessage(hRes);
    AfxMessageBox( sErr, MB_ICONEXCLAMATION | MB_OK);
}
#endif
